function [output] = mnscann(x,opts,ff)
%
% MNSCANN     Estimate family of level sets using kernel density estimation
%   and generate associated annotations
%
% USAGE
%
%   [output] = mnscann(xtrain,opts)
%
% INPUT
%
%   xtrain: d x n_train training data
%   opts: struct specifying parameters of learning algorithm
%
% Fields of "opts" are as follows: [default in brackets]
%
%     bandwidth: []
%       bandwidth of kernel; if not specified or nonpositive, selected by
%       minimum integrated volume (MIV) criterion 
%     n_grid: [2500]
%       number of grid points for estimating volume and for plotting 
%       set estimate in 2d; larger value means slower algorithm
%     supp_shape: [0]
%       0: construct mu on a bounding box (required for 2-d displays)
%       1: construct me on a bounding ball (required for skewed reference measure)
%     extreme: [0]
%       0: remove exteme outliers
%       1: don't
%     ref_measure: [0]
%       0: uniform
%       1: skewed (increasing mass toward periphery)
%     show_movie: [0] 
%       1: if d=2, create movie of level set estimates
%       0: don't display set estimate. 
%          note: displaying the estimate will slow down the program
%     show_contour: [0] 
%       1: if d=2, depict level set estimates by family of contours
%       0: don't display set estimate. 
%          note: displaying the estimate will slow down the program
%     show_iv: [0] 
%       1: show integrated volume curve as a function of bandwidth
%           generated by MIV bandwidth estimator
%       0: don't 
%     show_roc: [0]
%       1: show volume-mass curve for each bandwidth
%           evaluated by MIV bandwidth estimator
%       0: don't 
%     show_smooth: [0]
%       1: show the smoothed and unsmoothed ROCs used to estimate pi
%       0: don't
%     verbose: [0] verbose mode flag
%       1: print updates of different stages of the algorithm
%       0: don't
%
% OUTPUT:
%
% The struct "output" has these fields:
%
%   gammas          1 x n_train vector of estimates gamma values (1 - qvalue)
%   betas           1 x n_train vector of estimates beta values (1 - pvalue)
%   pi_est          estimate of fraction of outliers
%   vols            row vector of estimated volumes (extreme points excluded)
%   vols_raw        row vector of unsmoothed volumes (extreme points excluded)
%   masses          row vector of estimated Q-masses (extreme points excluded)
%   extreme         1 x n_train indicator vector, 1's indicate extreme outliers
%   dim             dimension of data
%   e_time          elapsed time 
%   bandwidth       estimated bandwidth using AUC criterion
%   coarse_hgrid    bandwidths considered -- coarse search
%   coarse_iv       integrated volumes of bandwidth candidates
%   fine_hgrid      bandwidths considered -- fine search
%   fine_iv         integrated volumes of bandwidth candidates
%   mov             movie of GQ/MV sets sorted by volume
%

%
% AUTHOR
%
% Clayton Scott, 2007-2009, University of Michigan
%
% All rights reserved

t0 = clock;

global xtrain xtest XTR XTS n_train n_test params d labels

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% initialize some variables
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

params = opts; % redefine as globally visible variable
d=size(x,1);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% set defaults, check for errors, preprocess, etc.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if ~isfield(params,'show_movie')
    params.show_movie=0;
end
if params.show_movie == 1    
    if (d ~=2)
        warning('Can only show movie for 2D data')
        params.show_movie=0;
    else
        params.supp_shape = 0;        
    end
end
if ~isfield(params,'show_contour')
    params.show_contour=0;
end
if params.show_contour == 1    
    if (d ~=2)
        warning('Can only show contours for 2D data')
        params.show_contour=0;
    end
end
if ~isfield(params,'supp_shape')
    params.supp_shape = 0;
end
if ~isfield(params,'ref_measure')
    params.ref_measure = 0;
end
if (params.show_movie | params.show_contour) & params.supp_shape
    error('Support must be square to show movies or contours')
end
if params.supp_shape == 0 & params.ref_measure > 0
    error('Support must be spherical to apply non-uniform reference measure')
end
if ~isfield(params,'extreme')
    params.extreme = 0;
end
if ~isfield(params,'n_grid')
    params.n_grid = 2500;
end
if ~isfield(params,'show_iv')
    params.show_iv = 0;
end
if ~isfield(params,'show_roc')
    params.show_roc = 0;
end
if ~isfield(params,'show_smooth')
    params.show_smooth = 0;
end
if ~isfield(params,'verbose')
    params.verbose = 0;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% identify extreme outliers and normalize rest of data to unit cube
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if params.supp_shape == 0 % box
    if params.extreme == 0
        x = prctbox(x,1,0); % assume no extreme outliers
    else
        x = prctbox(x,.98,.5); % preprocesses extreme outliers; 
    end
    output.extreme = (sum(x >= 0 & x <= 1)~=d); % points outside the unit hypercube are extreme outliers
else % ball
    if params.extreme == 0 
        x = prctball(x,1,0); % assume no extreme outliers        
    else
        x = prctball(x,.98,.1); % preprocesses extreme outliers; 
    end
    output.extreme = (sum(x.^2) > 1); % points outside the unit ball are extreme outliers
end

% effective training data
xtrain = x(:,find(~output.extreme));
n_train = size(xtrain,2);

%keyboard

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Set bandwidth of KDE level set estimator
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% generate uniform points for estimating volume
if d ==2 & params.supp_shape == 0 % regular grid to facilitate graphics
    n_side = floor(sqrt(params.n_grid));
    params.n_grid = n_side^2;
    x1=reshape(repmat(linspace(0,1,n_side),[n_side,1]),1,n_side^2);
    x2=repmat(linspace(0,1,n_side),[1 n_side]);
    xtest=[x1; x2];
else % random uniform sample
    if params.supp_shape == 0
        xtest = rand(d,params.n_grid);
    else
        xtest = randball(d,params.n_grid,params.ref_measure);
    end
end
n_test = size(xtest,2);

%if d == 2 & params.verbose == 1
%    scatter(xtest(1,:),xtest(2,:))
%    title('sample from anomaly distribution')
%end

if nargin < 3 | ff == 0
    % precompute squared distances, which are used repeatedly in kernel
    % evaluation
    if params.verbose == 1
        fprintf('Pre-computing pairwise train-train squared distances\n')
    end
    for i = 1:n_train
        XTR(i,:) = sum((repmat(xtrain(:,i),1,n_train)-xtrain).^2);
    end
    if params.verbose == 1
        fprintf('Pre-computing pairwise train-test squared distances\n')
    end
    for i = 1:n_test
        XTS(i,:) = sum((repmat(xtest(:,i),1,n_train)-xtrain).^2);
    end
    XTS = XTS';    
else
    switch ff
        case 1
            load flow_data/sq2and5
        case 2
            load flow_data/sq1thru5
    end
end    

% estimate bandwidth if not specified by user
if ~isfield(params,'bandwidth') | params.bandwidth <= 0
    if params.verbose == 1
        disp('Estimating bandwidth')
    end
    [h_opt,coarse_hgrid,coarse_iv,fine_hgrid,fine_iv] = kde_bandwidth_miv();
    output.coarse_hgrid = coarse_hgrid;
    output.coarse_iv = coarse_iv;
    output.fine_hgrid = fine_hgrid;
    output.fine_iv = fine_iv;
else
    h_opt = params.bandwidth;
end
output.bandwidth = h_opt; % chosen bandwidth

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% learn multiple levels sets, extract volumes and masses
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if params.verbose == 1
    disp('Computing volumes and masses of level sets')
end
[vols, masses] = kde_all_levels(h_opt);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Assemble outputs: gammas, betas, vols, masses, pi_est
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% smooth and concavify empirical ROC
output.vols_raw = vols;
if params.verbose == 1
    disp('Smoothing volumes')
end
[vols,dens_est,pi_est]=roc_lin_spline(vols,masses,params);

output.gammas(find(~output.extreme)) = min(pi_est*(1-vols)./(1-masses),1);
output.gammas(find(output.extreme)) = 1; % extreme outliers have gamma=1 by default

output.betas(find(~output.extreme)) = min((masses-pi_est*vols)/(1-pi_est),1);
output.betas(find(output.extreme)) = 1; % extreme outliers have beta=1 by default

output.vols = vols;
output.masses = masses;
output.pi_est = pi_est;
output.dim = d;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Form images and relevant summary statistics
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% generate movie of estimates
if params.show_movie==1
    [tmp,indx] = sort(masses);
    n_side = sqrt(n_test);
    figure
    for i=1:floor(n_train/5);
        clf

        % plot set
        set_map=labels(1+5*(i-1),:);
        set_im = reshape(set_map,[n_side,n_side]);
        lims = [0 1];
        imagesc(lims,lims,.4*set_im+.6,[0,1]);
        colormap(gray)
        axis xy
        hold on

        % plot data
        scatter(x(1,:),x(2,:),'o');        

        M(i) = getframe; % store movie frame
    end
    output.mov = M;
end

% generate contours of level sets
if params.show_contour == 1
    figure
    Z = sum(gauss_kern(XTS,h_opt,d))/n_train;
    contour(reshape(xtest(1,:),n_side,n_side),reshape(xtest(2,:),n_side,n_side),reshape(Z,n_side,n_side),20);
    drawnow
end

output.e_time = etime(clock,t0);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [h_opt,coarse_hgrid,coarse_iv,fine_hgrid,fine_iv] = kde_bandwidth_miv()
%
% select kde bandwidth by minimum integrated volume (= 1 - AUC)
% 

global xtrain n_train params d 

% local parameters
N_coarse = 20; % coarse bandwidth grid size
N_fine = 10; % fine bandwidth grid size
elo = -1.5; ehi = 1; % exponents defining range of search grid relative to scale of sample

% set up roc graph
if params.show_roc == 1
    roc_fig = figure;
end

% estimate scale of data for setting bandwidth grid
m = mean(xtrain,2); 
s = sqrt(sum(sum((xtrain - repmat(m,1,n_train)).^2))/(n_train*d));
se = log10(s); % exponent corresponding to data scale

% coarse grid search
coarse_hgrid = logspace(se+elo, se+ehi, N_coarse);
coarse_iv_min = Inf;
valid_h = 0; % set to one once we encounter an h that return nontrivial results
for i = 1:N_coarse
    
    h = coarse_hgrid(i);
    if params.verbose == 1
        fprintf('     Coarse iteration %d of %d\n',i,N_coarse)
    end
    
    % get the roc
    [vols, masses] = kde_all_levels(h);
    if max(vols) == 0 
        if valid_h == 1 % bandwidths beyond this point will only be worse
            coarse_iv(i:N_coarse) = Inf;
            break
        else
            coarse_iv(i) = Inf;
            continue
        end
    else
        valid_h == 1;
    end
    [sorted_masses,vol_idx]=sort(masses);
    sorted_vols=vols(vol_idx);
    coarse_iv(i) = sum(sorted_vols)/n_train;
    
    if params.show_roc == 1
        figure(roc_fig)
        clf
        plot(sorted_masses,sorted_vols,'.')
        ylabel('volume')
        xlabel('mass')
        title(['Integrated volume = ' num2str(coarse_iv(i))])
    end

    if coarse_iv(i) < coarse_iv_min
        coarse_iv_min = coarse_iv(i);
        coarse_h_opt = h;
    end
    drawnow
end
if params.show_iv == 1
    iv_fig = figure;
    semilogx(coarse_hgrid,coarse_iv,'o')
    drawnow
    hold on
end

% fine grid search in vicinity of coarse estimate
he = log10(coarse_h_opt);
fine_hgrid = logspace(he-(ehi-elo)/N_coarse, he+(ehi-elo)/N_coarse, N_fine);
fine_iv_min = coarse_iv_min;
h_opt = coarse_h_opt;
valid_h = 0; % set to one once we encounter an h that return nontrivial results
for i = 1:N_fine
    h = fine_hgrid(i);
    if params.verbose == 1
        fprintf('     Fine iteration %d of %d\n',i,N_fine)
    end
    
    % get the roc
    [vols, masses] = kde_all_levels(h);
    if max(vols) == 0 
        if valid_h == 1
            fine_iv(i:N_fine) = Inf;
            break
        else
            fine_iv(i) = Inf;
            continue
        end
    else
        valid_h == 1;
    end
    [sorted_masses,vol_idx]=sort(masses);
    sorted_vols=vols(vol_idx);
    fine_iv(i) = sum(sorted_vols)/n_train;

    if fine_iv(i) < fine_iv_min
        fine_iv_min = fine_iv(i);
        h_opt = fine_hgrid(i);
    end
    drawnow    
end

if params.show_iv == 1
    figure(iv_fig)
    semilogx(fine_hgrid,fine_iv)
    drawnow
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [vols, masses] = kde_all_levels(h)
%
% find all level sets
%
% h = targeted bandwidth for kde method
%

global XTR XTS n_train n_test params d labels

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% evaluate KDE at train and test points
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if 1 % CV: produces smoother IV curves with smaller minimums
    ftrain = (sum(gauss_kern(XTR,h,d)) - gauss_kern(diag(XTR)',h,d))/(n_train-1);
else % resub
    ftrain = sum(gauss_kern(XTR,h,d))/n_train; 
end
ftest = sum(gauss_kern(XTS,h,d))/n_train;

% initialize 
if params.show_movie == 1
    labels = ones(n_train,n_test); % n_train x n_test matrix of labels (0/1) assigned by each G_i
end

% compute volumes
[lam_grid0,data_idx] = sort(ftrain);
data_idx = fliplr(data_idx);
lam_diff = diff(lam_grid0);
lam_grid = lam_grid0(1:end-1) + lam_diff;
lam_grid = [lam_grid0(1)-(lam_grid0(2)-lam_grid0(1))/2 lam_grid lam_grid0(end)+(lam_grid0(end)-lam_grid0(end-1))/2];
nlam=length(lam_grid);

for i=1:nlam
    vols_sort0(nlam-i+1)=sum(ftest > lam_grid(i))/n_test;
    if params.show_movie == 1 & i < nlam
        labels(nlam-i+1,:) = (ftest < lam_grid(i));
    end
end
expec_num_anoms = n_train*0.1; % 0.1 = guess at pi
vols_sort = vols_sort0(1:end-1) + diff(vols_sort0)/2;
if vols_sort(end) == 0
    keyboard
end
vols_sort = vols_sort/(vols_sort(end)*(expec_num_anoms + 1)/expec_num_anoms); 
% normalize by volume of support, with bias-correction term

% assign masses
masses_sort = [0:n_train-1]/n_train + 1/(2*n_train);

% return volumes and masses to original order
[dummy,data_idx_rev] = sort(data_idx); % reverse mapping
masses = masses_sort(data_idx_rev); % return to original ordering
vols = vols_sort(data_idx_rev);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [phi] = gauss_kern(dist,h,d)
%
% evaluate gaussian kernel of dataset based on squared distances to kernel
% centers; 
%
% dist = matrix of squared distances
% h = bandwidth
% d = dimension
%

%phi = exp(-dist/(2*h^2))/((2*pi*h^2)^(d/2));
phi = exp(-dist/(2*h^2)); % constants unnecessary, and create underflow problem
